-- vertex
#version 460 core

layout(location = 0) in vec4 pos;

out vec2 tex;
out vec4 viewRay;

uniform mat4 proj;

void main(void)
{
	tex         = pos.zw;
	viewRay     = vec4 ( pos.x / proj [0][0], pos.y / proj [1][1], -1.0, 1.0 );
	gl_Position = vec4 ( pos.xy, 0.0, 1.0 );
}

-- fragment
#version 460 core

struct  Vertex
{
    float   f [6];
};

layout ( std430, binding = 0 ) buffer Vertices
{
    float   f [];
};

uniform usampler2D visMap;
uniform sampler2D  decalMap;

uniform mat4 mv;
uniform mat4 proj;
uniform int  mode;

in  vec2  tex;
out vec4  viewRay;
out vec4  color;

vec4    getPos ( int triangle, int no )
{
    int index = 3 * 6 * triangle + no * 6;
    
    return vec4 ( f[index + 0], f[index + 1], f[index + 2], f[index + 3] );
}

vec2    getTex ( int triangle, int no )
{
    int index = 3 * 6 * triangle + no * 6;
    
    return vec2 ( f[index + 4], f[index + 5] );
}

void main(void)
{
    vec2    winSize = vec2 ( textureSize ( visMap, 0 ) ); 
    vec2    ndc     = 2*tex - vec2 ( 1.0 );     // convert tex to NDC
    uint    id      = texture ( visMap, tex ).x;
    
    if ( id < 1 )	// skip empty space (with 0s)
        discard;
        
    id--;

        // get triangle transformed vertices 
    mat4    m    = proj * mv;
    vec4    v0   = m * getPos ( 0, 0 );
    vec4    v1   = m * getPos ( 0, 1 );
    vec4    v2   = m * getPos ( 0, 2 );
    vec3    w    = vec3 ( v0.w, v1.w, v2.w );
    vec3    invW = 1.0 / w;

        // compute NDC coordinates of triangle vertices
    vec2    p0 = v0.xy * invW.x;
    vec2    p1 = v1.xy * invW.y;
    vec2    p2 = v2.xy * invW.z;

        // compute 1/D
    float   invD = 1.0 / determinant ( mat2 ( p2 - p1, p0 - p2 ) );

        // compute barycentrics lambda of NDC point (x,y)
    vec3    lambda;

    lambda.x = (( p1.y - p2.y )*( ndc.x - p2.x ) + ( p2.x - p1.x ) * ( ndc.y - p2.y )) * invD;
    lambda.y = (( p2.y - p0.y )*( ndc.x - p2.x ) + ( p0.x - p2.x ) * ( ndc.y - p2.y )) * invD;
    lambda.z = 1.0 - lambda.x - lambda.y;

        // compute gradient
    vec3    lx = vec3 ( p1.y - p2.y, p2.y - p0.y, p0.y - p1.y ) * invD;
    vec3    ly = vec3 ( p2.x - p1.x, p0.x - p2.x, p1.x - p0.x ) * invD;

        // compute gradient of 1/w
    float   diwdx  = lx.x * invW.x + lx.y * invW.y + lx.z * invW.z;	// dot ( lx, invW )
    float   diwdy  = ly.x * invW.x + ly.y * invW.y + ly.z * invW.z;

        // compute w at (x,y) by linearly interpolating 1/w
    vec2    delta = ndc - p0;

        // now lerp 1/w to get w at ndc
    float   iwp = invW.x + delta.x * diwdx + delta.y * diwdy;  
    float   wp  = 1.0 / iwp;

        // convert to pixels
    lx *= 2.0 / winSize.x;
    ly *= 2.0 / winSize.y;

        // invert y (Vulkan only)
        //dldy = -dldy;

        // compute w(x+1,y) and w(x,y+1)
    float   wx1 = 1.0 / ( iwp + dot ( lx, invW ) );
    float   wy1 = 1.0 / ( iwp + dot ( ly, invW ) );

        // now compute lambda(x+1,y) and lambda(x,y+1)
    vec3    lx1 = wx1 * ( lambda * iwp + lx * invW );
    vec3    ly1 = wy1 * ( lambda * iwp + ly * invW );

        // and now get final lambda gradients
    vec3    ddx = lx1 - lambda;
    vec3    ddy = ly1 - lambda;

    ddx = ddx * invW;
    ddy = ddy * invW;

        // now do perspective-correct texture coordinates interpolation
    vec2    tx0 = getTex ( 0, 0 );
    vec2    tx1 = getTex ( 0, 1 );
    vec2    tx2 = getTex ( 0, 2 );
        
    float   den = dot ( lambda, invW );
    vec2    uv  = (tx0 * lambda.x * invW.x + tx1 * lambda.y * invW.y + tx2 * lambda.z * invW.z) / den;

        // compute gradients
    float dudx   = dot  ( vec3 ( tx0.x, tx1.x, tx2.x ), ddx );    
    float dudy   = dot  ( vec3 ( tx0.x, tx1.x, tx2.x ), ddy ); 
    float dvdx   = dot  ( vec3 ( tx0.y, tx1.y, tx2.y ), ddx );    
    float dvdy   = dot  ( vec3 ( tx0.y, tx1.y, tx2.y ), ddy ); 
    vec2  dtexdx = vec2 ( dudx, dvdx ) / den;
    vec2  dtexdy = vec2 ( dudy, dvdy ) / den;

        // find eye-space z
    float   z = (v0.z * lambda.x * invW.x + v1.z * lambda.y * invW.y + v2.z * lambda.z * invW.z ) / den;
    vec3    p = viewRay.xyz * z / viewRay.z;		// viewspace position

    color = textureGrad ( decalMap, uv, dtexdx, dtexdy );
}
